@@ -58,7 +58,6 @@ typedef struct {
 #define MAX_DYN_PREFIXES 4 /* the first 4 are dynamic */
 } sbcontext_t;
 
-static sbcontext_t sbcontext;
 static char *cached_env_vars[MAX_DYN_PREFIXES];
 bool sandbox_on = true;
 static bool sb_init = false;
@@ -73,10 +72,11 @@ static void init_env_entries(char ***, int *, const char *, const char *, int);
 /*
  * Initialize the shabang
  */
-
+#if 0
 __attribute__((destructor))
 void libsb_fini(void)
 {
+	/* let the kernel reap our resources -- it's faster anyways */
 	int x;
 
 	sb_init = false;
@@ -92,11 +92,12 @@ void libsb_fini(void)
 		clean_env_entries(&(sbcontext.prefixes[x]),
 				&(sbcontext.num_prefixes[x]));
 }
+#endif
 
 __attribute__((constructor))
 void libsb_init(void)
 {
-	int old_errno = errno;
+	save_errno();
 
 #ifdef SB_MEM_DEBUG
 	mtrace();
@@ -109,7 +110,7 @@ void libsb_init(void)
 
 //	sb_init = true;
 
-	errno = old_errno;
+	restore_errno();
 }
 
 static const char *sb_get_fd_dir(void)
@@ -125,7 +126,7 @@ static const char *sb_get_fd_dir(void)
 
 static const char *sb_get_cmdline(pid_t pid)
 {
-#if !defined(SANDBOX_PROC_1_CMDLINE)
+#if !defined(SANDBOX_PROC_1_CMDLINE) && !defined(SANDBOX_PROC_SELF_CMDLINE) && !defined(SANDBOX_PROC_dd_CMDLINE)
 # error "how do i access a proc's cmdline ?"
 #endif
 	static char path[256];
@@ -277,8 +278,10 @@ char *egetcwd(char *buf, size_t size)
 		char proc[20];
 		sprintf(proc, "/proc/%i/cwd", trace_pid);
 		ssize_t ret = readlink(proc, buf, size);
-		if (ret == -1)
+		if (ret == -1) {
+			errno = ESRCH;
 			return NULL;
+		}
 		buf[ret] = '\0';
 		return buf;
 	}
@@ -379,6 +382,34 @@ void sb_abort(void)
 	abort();
 }
 
+int sb_copy_file_to_fd(const char *file, int ofd)
+{
+	int ifd = sb_open(file, O_RDONLY, 0);
+	if (ifd == -1)
+		return -1;
+
+	size_t pagesz = getpagesize();
+	char *buf = xmalloc(pagesz);
+	while (1) {
+		size_t len = sb_read(ifd, buf, pagesz);
+		if (len == -1)
+			goto error;
+		else if (!len)
+			break;
+		size_t i;
+		for (i = 0; i < len; ++i)
+			if (!buf[i])
+				buf[i] = ' ';
+		SB_WRITE(ofd, buf, len, error);
+	}
+	sb_close(ifd);
+	free(buf);
+
+	return 0;
+ error:
+	return -1;
+}
+
 #define _SB_WRITE_STR(str) SB_WRITE(logfd, str, strlen(str), error)
 static bool write_logfile(const char *logfile, const char *func, const char *path,
                           const char *apath, const char *rpath, bool access)
@@ -435,26 +466,7 @@ static bool write_logfile(const char *logfile, const char *func, const char *pat
 
 	_SB_WRITE_STR("\nC: ");
 	const char *cmdline = sb_get_cmdline(trace_pid);
-	int cmdlinefd = sb_open(cmdline, O_RDONLY, 0);
-	if (cmdlinefd != -1) {
-		size_t pagesz = getpagesize();
-		char *buf = xmalloc(pagesz);
-		while (1) {
-			size_t len = sb_read(cmdlinefd, buf, pagesz);
-			if (len == -1) {
-				SB_EERROR("ISE:write_logfile ", "cmdlinefd read error\n");
-				break;
-			} else if (!len)
-				break;
-			size_t i;
-			for (i = 0; i < len; ++i)
-				if (!buf[i])
-					buf[i] = ' ';
-			SB_WRITE(logfd, buf, len, error);
-		}
-		sb_close(cmdlinefd);
-		free(buf);
-	} else {
+	if (sb_copy_file_to_fd(cmdline, logfd)) {
 		_SB_WRITE_STR("unable to read ");
 		_SB_WRITE_STR(cmdline);
 	}
@@ -906,10 +918,17 @@ static int check_syscall(sbcontext_t *sbcontext, int sb_nr, const char *func,
 		return 2;
 	}
 
+	/* Process went away while we were tracing it ... #264478 */
+	if (trace_pid && errno == ESRCH)
+		return 2;
+
 	/* If we get here, something bad happened */
 	SB_EERROR("ISE ", "%s(%s): %s\n"
 		"\tabs_path: %s\n" "\tres_path: %s\n",
 		func, file, strerror(errno), absolute_path, resolved_path);
+	const char *cmdline = sb_get_cmdline(trace_pid);
+	sb_copy_file_to_fd(cmdline, STDERR_FILENO);
+	sb_printf("\n%s\n", cmdline);
 	sb_abort();
 }
 
@@ -936,13 +955,10 @@ bool is_sandbox_on(void)
 	return result;
 }
 
-/* Need to protect the global sbcontext structure */
-static pthread_mutex_t sb_syscall_lock = PTHREAD_MUTEX_INITIALIZER;
-
 bool before_syscall(int dirfd, int sb_nr, const char *func, const char *file, int flags)
 {
 	int result;
-//	static sbcontext_t sbcontext;
+	static sbcontext_t sbcontext;
 	char at_file_buf[SB_PATH_MAX];
 
 	if (file == NULL || file[0] == '\0') {
@@ -984,7 +1000,8 @@ bool before_syscall(int dirfd, int sb_nr, const char *func, const char *file, in
 		file = at_file_buf;
 	}
 
-	pthread_mutex_lock(&sb_syscall_lock);
+	/* Need to protect the global sbcontext structure */
+	sb_lock();
 
 	if (!sb_init) {
 		init_context(&sbcontext);
@@ -1026,7 +1043,7 @@ bool before_syscall(int dirfd, int sb_nr, const char *func, const char *file, in
 
 	result = check_syscall(&sbcontext, sb_nr, func, file, flags);
 
-	pthread_mutex_unlock(&sb_syscall_lock);
+	sb_unlock();
 
 	if (0 == result) {
 		if ((NULL != getenv(ENV_SANDBOX_PID)) && (is_env_on(ENV_SANDBOX_ABORT)))
