@@ -11,7 +11,7 @@
 # define WRAPPER_ARGS_PROTO_FULL _WRAPPER_ARGS_PROTO, mode_t mode
 # define WRAPPER_ARGS dirfd, pathname, flags
 # define WRAPPER_ARGS_FULL WRAPPER_ARGS, mode
-# define WRAPPER_SAFE() FUNCTION_SANDBOX_SAFE_OPEN_INT_AT(dirfd, pathname, flags)
+# define WRAPPER_SAFE() SB_SAFE_OPEN_INT_AT(dirfd, pathname, flags)
 # define USE_AT 1
 #else
 # define USE_AT 0
