@@ -11,7 +11,6 @@
 	char *file = s; \
 	\
 	s = argv[i++]; \
-	mode_t mode; \
-	sscanf(s, "%i", &mode);
+	mode_t mode = sscanf_mode_t(s);
 
 #include "test-skel-0.c"
