@@ -8,7 +8,7 @@
 
 #define process_args() \
 	s = argv[i++]; \
-	int dirfd = atoi(s); \
+	int dirfd = at_get_fd(s); \
 	\
 	s = argv[i++]; \
 	char *file = s; \
@@ -17,7 +17,6 @@
 	const struct timespec *times = NULL; \
 	\
 	s = argv[i++]; \
-	int flags; \
-	sscanf(s, "%i", &flags);
+	int flags = at_get_flags(s);
 
 #include "test-skel-0.c"
